"""Module containing the ApplyGraph class"""
import concurrent.futures
from typing import List, Dict, Set

import networkx

from terrawrap.utils.graph import find_source_nodes
from terrawrap.models.graph_entry import GraphEntry, NoOpGraphEntry, Entry


class ApplyGraph:
    """Class for representing an Apply Graph."""

    def __init__(self, command: str, graph: networkx.DiGraph, post_graph: List[str], prefix: str):
        """
        :param command: The Terraform command that this pipeline should execute.
        :param graph: The graph to be executed.
        :param post_graph: The list of items to be executed after the graph has been run.
        :param prefix: The prefix an item must match to be applied.
        """
        self.command = command
        self.graph = graph
        self.graph_dict: Dict[str, Entry] = {}
        self.post_graph = post_graph
        self.prefix = prefix
        self.not_applied: Set[str] = set()
        self.applied: Set[str] = set()
        self.failures: List[str] = []

    # pylint: disable=too-many-locals
    def execute_graph(self, num_parallel: int = 4, debug: bool = False, print_only_changes: bool = False):
        """
        Function for executing the graph. Will execute in parallel, up to the limit given in num_parallel.
        :param num_parallel: The number of pipeline entries to run in parallel.
        :param debug: True if Terraform debugging should be turned on.
        :param print_only_changes: True if only directories which contained changes should be printed.
        """
        sources = find_source_nodes(self.graph)
        futures_to_paths = {}

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_parallel) as executor:
            for source in sources:
                entry = self._get_or_create_entry(source)

                future = executor.submit(entry.execute, self.command, debug=debug)
                futures_to_paths[future] = entry.path

            for future in concurrent.futures.as_completed(futures_to_paths):
                path = futures_to_paths[future]
                exit_code, stdout, changes_detected = future.result()

                if stdout and print_only_changes and not changes_detected:
                    stdout = ["No changes detected.\n"]

                print(f"Output for {path}:\n\n{''.join(stdout).strip()}\n")

                if exit_code != 0:
                    self.failures.append(path)

                successors = list(self.graph.successors(path))
                if successors:
                    self.recursive_executor(executor, successors, num_parallel, debug, print_only_changes)

        for node in self.graph:
            item = self.graph_dict.get(node)
            if not item:
                self.not_applied.add(node)
            elif item.state == "no-op":
                self.not_applied.add(item.path)
            else:
                self.applied.add(node)

    def recursive_executor(
            self,
            executor: concurrent.futures.Executor,
            successors: List[str],
            num_parallel: int,
            debug: bool,
            print_only_changes: bool):
        """
        Helper function for executing graph entries recursively
        :param executor: The Executor to use. See concurrent.futures.Executor.
        :param successors: A list of successors to be executed from the previous call
        :param num_parallel: The number of pipeline entries to run in parallel.
        :param debug: True if Terraform debugging should be turned on.
        :param print_only_changes: True if only directories which contained changes should be printed.
        """
        futures_to_paths = {}

        for node in successors:
            entry = self._get_or_create_entry(node)
            if entry.state != "Pending":
                continue
            if not self._can_be_applied(entry):
                continue

            future = executor.submit(entry.execute, self.command, debug=debug)
            futures_to_paths[future] = entry.path

        for future in concurrent.futures.as_completed(futures_to_paths):
            path = futures_to_paths[future]
            exit_code, stdout, changes_detected = future.result()

            if stdout and print_only_changes and not changes_detected:
                stdout = ["No changes detected.\n"]

            print(f"Output for {path}:\n\n{''.join(stdout).strip()}\n")
            if exit_code != 0:
                self.failures.append(path)

            next_successors = list(self.graph.successors(path))
            if next_successors:
                self.recursive_executor(
                    executor, next_successors, num_parallel, debug, print_only_changes
                )

    def execute_post_graph(
            self,
            num_parallel: int = 4,
            debug: bool = False,
            print_only_changes: bool = False):
        """
        Function for executing entries not in the graph in parallel.
        :param num_parallel: The number of pipeline entries to run in parallel.
        :param debug: True if Terraform debugging should be turned on.
        :param print_only_changes: True if only directories which contained changes should be printed.
        """
        futures_to_paths = {}

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_parallel) as executor:
            for node in self.post_graph:
                entry = self._get_or_create_entry(node)

                future = executor.submit(entry.execute, self.command, debug=debug)
                futures_to_paths[future] = entry.path

            for future in concurrent.futures.as_completed(futures_to_paths):

                path = futures_to_paths[future]
                if self.graph_dict[path].state != "no-op":

                    exit_code, stdout, changes_detected = future.result()

                    if print_only_changes and not changes_detected:
                        stdout = ["No changes detected.\n"]

                    print(f"Output for {path}:\n\n{''.join(stdout).strip()}\n")

                    if exit_code != 0:
                        self.failures.append(path)

        for node in self.post_graph:
            item = self.graph_dict.get(node)
            if not item:
                self.not_applied.add(node)
            elif item.state == "no-op":
                self.not_applied.add(item.path)
            else:
                self.applied.add(node)

    def _can_be_applied(self, entry: GraphEntry):
        """
        Checks if an entry can be applied.
        :param entry: The entry to be tested.
        :return: A boolean False if the entry cannot be applied otherwise True.
        """
        if entry:
            path = entry.path
            predecessors = list(self.graph.predecessors(path))

            for predecessor in predecessors:
                pred_entry = self.graph_dict.get(predecessor)

                if pred_entry:
                    if pred_entry.state not in ("Success", "no-op"):
                        return False
                else:
                    return False

        else:
            return False

        return True

    def _get_or_create_entry(self, node: str):
        """
        Gets an entry from the graph dictionary or create it if it does not exist
        :param node: The node used to fetch the graph entry
        :return: The graph entry
        """
        if self.graph_dict.get(node):
            entry = self.graph_dict.get(node)
        else:
            if node.startswith(self.prefix):
                entry = GraphEntry(node, [])
            else:
                entry = NoOpGraphEntry(node, [])
            self.graph_dict[node] = entry
        return entry



"""Utilities for working with Terraform variables"""
import concurrent.futures
import os
from collections import defaultdict, namedtuple
from typing import Dict, Set, Tuple, Union

import hcl2
from lark import Token
from networkx import DiGraph

Variable = namedtuple('Variable', ['name', 'value'])


def get_auto_vars(root_directory: str) -> Dict[str, Set[Variable]]:
    """
    Recursively scan a directory and find all variables that are being exposed via tfvars files
    :param root_directory: directory where to start search
    :return:
    """
    auto_vars: Dict[str, Set[Variable]] = defaultdict(set)
    # pylint: disable=unused-variable
    for current_dir, dirs, files in os.walk(root_directory, followlinks=True):
        if '.terraform' in current_dir:
            continue

        vars_files = [file_name for file_name in files if file_name.endswith('.tfvars')]
        for file_name in vars_files:
            with open(current_dir + '/' + file_name, 'r', encoding='utf-8') as file:
                variables = hcl2.load(file)
                for key, value in variables.items():
                    auto_vars[os.path.join(current_dir, file_name)].add(Variable(key, _make_hashable(value)))

    return dict(auto_vars)


def _make_hashable(input_value):
    if isinstance(input_value, list):
        return tuple(_make_hashable(item) for item in input_value)
    if isinstance(input_value, dict):
        return tuple((_make_hashable(key), _make_hashable(value)) for key, value in input_value.items())
    if isinstance(input_value, Token):
        return str(input_value)
    return input_value


def get_nondefault_variables_for_file(file_path: str) -> Set[str]:
    """
    Find all variables missing default values that are declared in a terraform file
    :param file_path: a terraform file
    :return: Set of variable names declared in the file
    """
    variables = set()
    with open(file_path, 'r', encoding='utf-8') as file:
        tf_info = hcl2.load(file)
        for variable in tf_info.get('variable', []):
            for variable_name, var_config in variable.items():
                if not var_config.get('default'):
                    variables.add(variable_name)

    return variables


def get_source_for_variable(usage_directory: str, var_name: str, vars_map: Dict[str, Set[Variable]]) \
        -> Union[None, str]:
    """
    Find which auto tfvars file terraform will use to get a variable value from
    :param usage_directory: A directory with terraform config that's using a variable
    :param var_name: Name of a variable that needs to get a value via auto tfvars
    :param vars_map: Dict of tfvars files and the vars they provided returned by `get_auto_vars`
    :return: file path of tfvars file with the value for the variable
    """
    # find which var files have the var we want and are in a directory above the one we are in
    possible_sources = [
        file
        for file, var_info in vars_map.items()
        if any(var[0] == var_name for var in var_info) and usage_directory.startswith(os.path.dirname(file))
    ]

    if not possible_sources:
        return None

    # sort the list by how deeply nested it is
    possible_sources = sorted(possible_sources, key=lambda s: len(s.split('/')))

    # the most deeply nested one is the one closest to where we are using the variable
    # which is the one that would get used
    return possible_sources[-1]


def get_auto_var_usage_graph(root_directory: str) -> DiGraph:
    """
    Recursively scan a directory to build a graph of auto tfvars files and the directories
    that depend on them
    :param root_directory: directory where to start the search
    :return:
    """
    graph = DiGraph()
    auto_vars = get_auto_vars(root_directory)

    future_list = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
        # pylint: disable=unused-variable
        for current_dir, dirs, files in os.walk(root_directory, followlinks=True):
            if '.terraform' in current_dir or '.git' in current_dir:
                continue

            for file in files:
                if not file.endswith('.tf'):
                    continue

                future = executor.submit(_collect_variable_usages, current_dir, file, auto_vars)
                future_list.append(future)

        for future in concurrent.futures.as_completed(future_list):
            directory, var_sources = future.result()
            for var_source in var_sources:

                if var_source not in graph.nodes:
                    graph.add_node(var_source)

                if directory not in graph.nodes:
                    graph.add_node(directory)

                graph.add_edge(var_source, directory)

    return graph


def _collect_variable_usages(current_dir: str, file: str, auto_vars: Dict[str, Set[Variable]]) \
        -> Tuple[str, Set[str]]:
    var_sources = set()
    for variable in get_nondefault_variables_for_file(os.path.join(current_dir, file)):
        var_source = get_source_for_variable(current_dir, variable, auto_vars)
        if var_source:
            var_sources.add(var_source)
    return current_dir, var_sources



"""
Container for terrawrap exceptions
"""


class NotTerraformConfigDirectory(RuntimeError):
    """Error raised when processing a directory that contains no .tf config files"""


class NoDependency(Exception):
    """Error raised when processing a directory that contains .tf_wrapper config files with no dependency"""



"""Module for containing Graph Entries"""
from abc import ABC, abstractmethod

import os
from typing import List, Tuple

from terrawrap.utils.cli import execute_command
from terrawrap.utils.config import (
    find_wrapper_config_files,
    parse_wrapper_configs,
    resolve_envvars,
)
from terrawrap.utils.path import get_absolute_path


class Entry(ABC):
    """Abstract Graph Entry class"""
    path = ""
    state = ""

    @abstractmethod
    def execute(self, operation: str, debug: bool = False) -> Tuple[int, List[str], bool]:
        """Execute a graph entry"""
        return 0, [], False


class NoOpGraphEntry(Entry):
    """NoOp Graph Entry class. Use this for graph entries that should not be executed"""
    def __init__(self, path: str, variables: List[str]):
        self.path = path
        self.variables = variables
        self.state = "no-op"

    def execute(self, operation: str, debug: bool = False) -> Tuple[int, List[str], bool]:
        print(f"Skipping execute for {self.path} {operation} ...")
        return 0, [], False


class GraphEntry(Entry):
    """Class representing a Graph Entry"""

    def __init__(self, path: str, variables: List[str]):
        """
        :param path: The path to the Terraform configuration files to execute.
        :param variables: Any additional variables to set alongside the Terraform command.
        """
        self.path = path
        self.abs_path = get_absolute_path(path=path)
        wrapper_config_files = find_wrapper_config_files(self.abs_path)
        self.wrapper_config = parse_wrapper_configs(wrapper_config_files)
        self.envvars = resolve_envvars(self.wrapper_config.envvars)
        self.variables = variables
        self.state = "Pending"

    # pylint: disable=too-many-locals
    def execute(self, operation: str, debug: bool = False) -> Tuple[int, List[str], bool]:
        """
        Function for executing this Graph Entry.
        :param operation: The Terraform operation to execute. IE: apply, plan
        :param debug: True if Terraform debug info should be printed.
        :return: A tuple of the exit code, output of the command, and whether changes were detected.
        """
        print(f"Executing {self.abs_path} {operation} ...")
        self.state = "Executing"
        command_env = os.environ.copy()
        command_env.update(self.envvars)

        if debug:
            command_env["TF_LOG"] = "DEBUG"

        # We're using --no-resolve-envvars here because we've already resolved the environment variables in
        # the constructor. We are then passing in those environment variables explicitly in the
        # execute_command call below.
        base_args = ["tf", "--no-resolve-envvars", self.abs_path]
        init_args = base_args + ["init"] + self.variables
        operation_args = base_args + [operation] + self.variables

        init_exit_code, init_stdout = execute_command(
            init_args,
            print_output=False,
            capture_stderr=True,
            env=command_env,
            audit_api_url=self.wrapper_config.audit_api_url,
            cwd=self.abs_path
        )
        if init_exit_code != 0:
            self.state = "Failed"
            return init_exit_code, init_stdout, True

        shell = False
        if operation in ["apply", "destroy"]:
            operation_args = ["yes", "yes", "|"] + operation_args
            shell = True

        operation_exit_code, operation_stdout = execute_command(
            " ".join(operation_args) if shell else operation_args,
            print_output=False,
            capture_stderr=True,
            env=command_env,
            shell=shell,
            audit_api_url=self.wrapper_config.audit_api_url,
            cwd=self.path
        )

        if operation_exit_code == 0:
            self.state = "Success"
        else:
            self.state = "Failed"

        changes_detected = True
        if any("Resources: 0 added, 0 changed, 0 destroyed" in line for line in operation_stdout):
            changes_detected = False

        print(f"\nFinished executing {self.abs_path} {operation} ...")
        return (
            operation_exit_code,
            init_stdout + ["\n"] + operation_stdout,
            changes_detected,
        )



"""Module for containing convenience functions around path manipulation"""
import os
import re
import subprocess
from collections import defaultdict
from typing import Dict, Set

from networkx import DiGraph

GIT_REPO_REGEX = r"URL.*/([\w-]*)(?:\.git)?"


def get_absolute_path(path: str, root_dir: str = None) -> str:
    """
    Convenience function for determining the full path to a file or directory.
    A RuntimeError will be raised if the given path does not appear to point to either a file or a directory.
    :param path: The path. Can be either relative from the cwd or an absolute path.
    :param root_dir: The root directory to use instead of the cwd.
    :return: An absolute path for the given path.
    """
    if os.path.isabs(path):
        path = os.path.abspath(path)
    else:
        path = os.path.abspath(os.path.join(root_dir or os.getcwd(), path))

    return path


def get_symlinks(directory: str) -> Dict[str, Set[str]]:
    """
    Recursively walk a directory and return a dict of all symlinks
    :param directory:
    :return: dict of symlink source to set of paths that link to that source
    """
    links: Dict[str, Set[str]] = defaultdict(set)
    # pylint: disable=unused-variable
    for current_dir, dirs, files in os.walk(directory, followlinks=True):
        if '.terraform' in current_dir:
            continue

        if os.path.islink(current_dir):
            link_source = os.path.join(os.path.dirname(current_dir), os.readlink(current_dir))
            links[os.path.normpath(link_source)].add(os.path.normpath(current_dir))

    return dict(links)


def get_file_graph(directory: str) -> DiGraph:
    """
    Recursively walk a directory and return a graph of all files and symlinks
    :param directory:
    :return: graph of file names to parent directories and symlink destinations
    """
    graph = DiGraph()
    for current_dir, dirs, files in os.walk(directory):
        if '.terraform' in current_dir or '.git' in current_dir:
            continue

        if current_dir not in graph.nodes:
            graph.add_node(current_dir)

        # for every file in a dir, create a node and an edge pointing to the parent dir
        # also create an edge for symlinks to symlink source
        for path in files:
            norm_path = os.path.normpath(os.path.join(current_dir, path))
            if norm_path not in graph.nodes:
                graph.add_node(norm_path)

            graph.add_edge(norm_path, current_dir)

            if os.path.islink(norm_path):
                link_source = os.path.normpath(os.path.join(current_dir, os.readlink(norm_path)))

                if link_source not in graph.nodes:
                    graph.add_node(link_source)

                graph.add_edge(link_source, norm_path)

        # handle dirs the same way as files but don't create a node/edge for them unless they are a symlink
        for path in dirs:
            norm_path = os.path.normpath(os.path.join(current_dir, path))
            if os.path.islink(norm_path):
                link_source = os.path.normpath(os.path.join(current_dir, os.readlink(norm_path)))

                if link_source not in graph.nodes:
                    graph.add_node(link_source)

                graph.add_edge(link_source, norm_path)
    return graph


def calc_repo_path(path: str) -> str:
    """
    Convenience function for taking an absolute path to a TF directory and returning the path to that
    directory relative to the repo.
    :param path: The absolute path to a TF directory.
    :return: New path to the TF directory
    """
    byte_output = subprocess.check_output(["git", "remote", "show", "origin", "-n"], cwd=path)
    output = byte_output.decode("utf-8", errors="replace")
    match = re.search(GIT_REPO_REGEX, output)
    if match:
        repo_name = match.group(1)
    else:
        raise RuntimeError("Could not determine git repo name, are we in a git repo?")

    return f'{repo_name}/{path[path.index("/config") + 1:]}'



"""File Download Utils"""

import os
import platform
import stat
from typing import Dict, Tuple, Optional
from urllib.parse import urlparse

import boto3
import requests
from botocore.exceptions import ClientError
from filelock import FileLock


class FileDownloadFailed(RuntimeError):
    """Error raised when failing to download a file"""


class PluginDownload:
    """Utility for downloading plugins"""

    def __init__(self, s3_client):
        self.s3_client = s3_client or boto3.client('s3')

    def download_plugins(self, plugin_paths: Dict[str, str]):
        """
        Download a set of Terraform plugins to the user's home directory
        :param plugin_paths: A dictionary of plugin names and URLs where to download them
        """
        for name, path in plugin_paths.items():
            home = os.path.expanduser("~")
            plugin_directory = os.path.join(home, '.terraform.d/plugins')
            os.makedirs(plugin_directory, exist_ok=True)

            file_path = os.path.join(plugin_directory, name)

            system = platform.system()
            machine = platform.machine()
            path_with_platform = f'{path}/{system}/{machine}'

            lock_path = f'{file_path}.{"lock"}'
            # It seems that this is the name of the abstract class, but also an alias for the proper class, so
            # this warning is not relevant for us.
            # pylint: disable=abstract-class-instantiated
            lock = FileLock(lock_path, timeout=600)
            # use a lock to prevent conflicts writing the file if running this command in parallel
            with lock:
                try:
                    self._download_file(path_with_platform, file_path)
                except FileDownloadFailed:
                    print(f'Unable to get plugin from {path_with_platform}. Attempting {path} instead')
                    self._download_file(path, file_path)

    def _download_file(self, url: str, file_path: str):
        """
        Download from a url and save it to a file

        :param url: URL of file to download
        :param file_path: Path where to save file
        """
        # get the etag from the etag file if it exists
        etag = None
        etag_path = f'{file_path}.{"etag"}'
        if os.path.isfile(etag_path) and os.path.isfile(file_path):
            with(open(etag_path, 'r', encoding='utf-8')) as etag_file:
                etag = etag_file.read()

        download_info = self._get_file_content(url, etag)

        if download_info:
            content = download_info[0]
            etag = download_info[1]

            with(open(file_path, 'wb', encoding='utf-8')) as out_file:
                out_file.write(content)

            mode = os.stat(file_path)
            os.chmod(file_path, mode.st_mode | stat.S_IEXEC)

            if etag:
                # AWS returns the etag surrounded by quotes
                # remove them if that happens
                if etag[0] == '"':
                    etag = etag[1:-1]
                with(open(etag_path, 'w', encoding='utf-8')) as etag_file:
                    etag_file.write(etag)

    def _get_file_content(self, url: str, etag: Optional[str]) -> Optional[Tuple[bytes, Optional[str]]]:
        """
        Download a file from either S3 or HTTP

        :param url: URL where to download the file from
        :param etag: etag value to use for caching
        :return: File content and the file's etag. Will return None if the file is already cached
        """
        print(f'Downloading {url}')

        parsed_url = urlparse(url)

        if parsed_url.scheme in ('http', 'https'):
            return self._get_http_content(url, etag)

        if parsed_url.scheme == 's3':
            return self._get_s3_content(url, etag)

        raise RuntimeError('Invalid file download scheme. URL must start with one of (http, https, s3')

    def _get_http_content(self, url: str, etag: Optional[str]) -> Optional[Tuple[bytes, Optional[str]]]:
        """Download a file over HTTP/HTTPS"""
        headers = {}

        # get the file. send etag header to avoid downloading file if we already have the same version
        if etag:
            headers['If-None-Match'] = etag

        try:
            response = requests.get(url, headers=headers)
            response.raise_for_status()

            if response.status_code == 304:
                return None

            return response.content, response.headers.get('etag')
        except requests.HTTPError as exception:
            raise FileDownloadFailed() from exception

    def _get_s3_content(self, url: str, etag: Optional[str]) -> Optional[Tuple[bytes, Optional[str]]]:
        """Download a file from S3 using the AWS SDK"""
        parsed_url = urlparse(url)

        args = {
            'Bucket': parsed_url.hostname,
            'Key': parsed_url.path[1:]  # remove leading slash in path
        }

        if etag:
            args['IfNoneMatch'] = etag

        try:
            response = self.s3_client.get_object(**args)
            return response['Body'].read(), response['ETag']
        except ClientError as ex:
            if ex.response['Error']['Code'] == '304':
                return None
            if ex.response['Error']['Code'] == 'NoSuchKey':
                raise FileDownloadFailed() from ex
            raise ex



"""Module for containing CLI convenience functions"""
from __future__ import print_function

import getpass
import logging
import subprocess
import tempfile

from typing import List, Tuple, Union

import requests

from amplify_aws_utils.resource_helper import Jitter

from terrawrap.utils.git_utils import get_git_root

logger = logging.getLogger(__name__)

MAX_RETRIES = 5
RETRIABLE_ERRORS = [
    'RequestError: send request failed',
    'unexpected EOF',
    'Throttling',
    'timeout while waiting for state',
    'ServiceUnavailable: Service Unavailable',
    'failed to decode query XML error response',
    'connection reset',
    'Connection reset',
    'Please try again.',
    'Client.Timeout exceeded',
    'Request limit for operation',
    'try again later',
    'handshake timeout',
    'SSL_ERROR_SYSCALL',
    'ConditionalCheckFailedException',
    'Api Rate Limit Exceeded',
    'TooManyUpdates',
]


def execute_command(
        args: Union[List[str], str],
        *pargs,
        print_output: bool = True,
        capture_stderr: bool = True,
        print_command: bool = False,
        retry: bool = False,
        timeout: int = 15 * 60,
        audit_api_url: str = None,
        **kwargs
) -> Tuple[int, List[str]]:
    """
    Convenience function for executing a given command and optionally printing the output.
    :param args: List of arguments to execute, or a single string.
    :param pargs: Any additional positional arguments to Popen.
    :param print_output: True if the output of the command should be printed immediately. Defaults to True.
    :param capture_stderr: True if stderr should be captured. Defaults to True.
    :param print_command: True if the command should be printed before executing. Defaults to False.
    :param timeout: Max amount of time to keep retrying to execute command. Defaults to 15 minutes.
    :param retry: Retry a number of times if network errors. Defaults to False.
    :param audit_api_url: Audit API URL to submit POST request. Defaults to None so no data is sent.
    :param kwargs: Any additional keyword arguments to Popen.
    :return: A tuple of the exit code and output of the command.
    """
    try_count = 0

    # It's possible for an envvar to be set to none, so exclude those envvars.
    if 'env' in kwargs:
        kwargs['env'] = {
            key: value
            for key, value in kwargs['env'].items()
            if value is not None
        }

    jitter = Jitter()
    time_passed = 0
    exit_code = 0
    stdout: List[str] = []
    while try_count < MAX_RETRIES if retry else 1:
        exit_code, stdout = _execute_command(
            args,
            print_output,
            capture_stderr,
            print_command,
            *pargs,
            **kwargs,
        )

        try_count += 1

        network_errors = _get_retriable_errors(stdout)
        if exit_code != 0 and network_errors and retry:
            logger.warning('Found network errors while running %s command: %s', args, network_errors)
        else:
            # The command either succeeded or failed with a non network error. don't retry
            break

        if time_passed >= timeout:
            raise TimeoutError(f'Timed out retrying {args} command')

        time_passed = jitter.backoff()

    if audit_api_url and kwargs['cwd']:
        _post_to_audit_api_url(audit_api_url, kwargs['cwd'], exit_code, stdout)

    return exit_code, stdout


def _execute_command(
        args: Union[List[str], str],
        print_output: bool,
        capture_stderr: bool,
        print_command: bool,
        *pargs,
        **kwargs
) -> Tuple[int, List[str]]:
    """
    Private function for executing a given command and optionally printing the output.
    :param args: List of arguments to execute, or a single string.
    :param print_output: True if the output of the command should be printed immediately. Defaults to True.
    :param capture_stderr: True if stderr should be captured. Defaults to True.
    :param print_command: True if the command should be printed before executing. Defaults to False.
    :param pargs: Any additional positional arguments to Popen.
    :param kwargs: Any additional keyword arguments to Popen.
    :return: A tuple of the exit code and output of the command.
    """
    stdout_write, stdout_path = tempfile.mkstemp()
    with open(stdout_path, "rb") as stdout_read, open('/dev/null', 'w', encoding='utf-8') as dev_null:

        if print_command:
            print(f"Executing: {' '.join(args)}")

        kwargs['stdout'] = stdout_write
        kwargs['stderr'] = stdout_write if capture_stderr else dev_null

        # pylint: disable=consider-using-with
        process = subprocess.Popen(
            args,
            *pargs,
            **kwargs
        )

        while True:
            output = stdout_read.read(1).decode(errors="replace")

            if output == '' and process.poll() is not None:
                break

            if print_output and output:
                print(output, end="", flush=True)

        exit_code = process.poll()

        stdout_read.seek(0)
        stdout = [line.decode(errors="replace") for line in stdout_read.readlines()]

        # ignoring mypy error below because it thinks exit_code can sometimes be None
        # we know that will never be the case because the above While loop will keep looping forever
        # until exit_code is not None
        return exit_code, stdout  # type: ignore


def _get_retriable_errors(out: List[str]) -> List[str]:
    """Filter line output for retriable errors"""
    return [
        line for line in out
        if any(error in line for error in RETRIABLE_ERRORS)
    ]


def _post_to_audit_api_url(audit_api_url: str, path: str, exit_code: int, stdout: List[str]):
    root = get_git_root(path)
    path = path.replace(root, '')

    status = 'SUCCESS' if exit_code == 0 else 'FAILED'

    logger.info('Getpass.getuser() will be used to grab the user running path: %s', path)

    user = getpass.getuser()

    logger.info('Attempting to send data to Audit API: %s run by %s(%s)', path, user, status)

    try:
        requests.post(
            audit_api_url, json={
                'directory': path,
                'status': status,
                'run_by': user,
                'output': stdout
            })
        logger.info('Successfully posted data to provided url: %s', audit_api_url)
    except requests.exceptions.RequestException:
        logger.error("Unable to post data to provided url: %s", audit_api_url)
